;;; bbdb-expire.el --- expiry and expire-proof entries for the BBDB

;;; Copyright (C) 2000 -- 2002 Nix <nix@esperi.demon.co.uk>.

;; Author: Nix <nix@esperi.demon.co.uk>
;; Created: 2000-09-17
;; Last modified: 2002-01-28
;; Keywords: mail news
;; Version: $Revision: 1.5 $

;; This file is not part of XEmacs, or GNU Emacs.

;; This library is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; It is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
;; License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this library; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;; 02111-1307, USA.

;;; Commentary:

;; This file implements expiry of BBDB databases, to stop them getting too
;; fantastically bloated. Only specific entries are retained across expiry.
;;
;; Each entry is passed through each function in the
;; `bbdb-expire-expiry-functions' in turn; if any return t it is a candicate
;; for expiry; but if any of the functions in the
;; `bbdb-expire-preservation-functions' return t, it is not expired after all.
;; (This scheme allows functions to say `I don't care' as well as `expire
;; me now' and `do not expire me'.)
;;
;; One function is on `bbdb-expire-expiry-functions' by default;
;; `bbdb-expire-record-old-p'.
;; Two functions are on `bbdb-expire-preservation-functions' by default;
;; `bbdb-expire-field-permanent-p' and `bbdb-expire-field-notes-p'.
;;
;; Together, these ensure that old records without a `notes' or a
;; `permanent' field are expired; the field-checking functions are
;; generated by a macro, `bbdb-expire-field-foo-p'.

;; To use:

;; First of all, after you load this package but before you do anything else
;; with the BBDB, do a `bbdb-expire-add-last-access-time'. This attaches the
;; last access time field to all the BBDB entries, and assumes they were
;; last accessed *right now*. You never need to do this again, except if
;; you stop using this package and want to restart later on.

;; WARNING: If you forget to do this, only entries that you have accessed,
;; and then left for more than `bbdb-expire-this-old' days, will be expired;
;; entries that have been left since before you used this package will be
;; left alone, which is suboptimal.

;; (You do not need to do this if you remove `bbdb-expire-record-old-p'
;; from the `bbdb-expire-expiry-functions', but that leaves this package
;; mostly devoid of function...)

;; To expire the BBDB at any time, run `bbdb-expire-bbdb'.

;; Everything else happens automagically.

;; To do:

;; Currently, `bbdb-expire-bbdb' is not automatically called.
;; The ideal would be to call it once per session, as BBDB initializes, but
;; the only obvious place to do that is the `bbdb-after-read-db-hook',
;; but as you cannot call `bbdb-records' from that hook, this is impossible.
;; If you have any idea where I can hang it, please say!

;; (perhaps) Fully integrate into the BBDB. No matter what, `permanent' and
;; `last-access' should remain user fields (i.e. this should not trigger a
;; new revision of the BBDB file format), because those fields are updated by
;; optional predicate functions that can be removed and/or replaced with
;; anything else the user desires.

;; Add customize support.

;; This probably does not work with GNU Emacs; I don't currently have a copy
;; so it is hard to test it there, and there are probably unintentional
;; dependencies on XEmacs.
;;
;; If you want to use this on GNU Emacs, and it doesn't work, please tell me
;; what is missing or broken for GNU Emacs support so I can fix it, as I
;; would like it to work there as well.

;;; Requirements:

(require 'bbdb)
(require 'bbdb-com)
(require 'bbdb-hooks)
(require 'time-date)

(require 'cl)

;;; User-configurable variables:

(defvar bbdb-expire-this-old 90
  "Expire entries in the BBDB that have not been referenced for this many days.
Used by the `bbdb-expire-record-old-p' function.")

(defvar bbdb-expire-expiry-functions '(bbdb-expire-record-old-p)
  "Functions to call to determine if a record is expirable or not.
Each function is called with one parameter, the record to test for expirable
status; it should return t if the record is eligible for expiry.")

(defvar bbdb-expire-preservation-functions '(bbdb-expire-field-permanent-p
                                             bbdb-expire-field-notes-p)
  "Functions to call to determine if a record should be preserved from expiry.
Each function is called with one parameter, the record to test for preservable
status; it should return t if the record should never be expired.")

(defvar bbdb-expire-suppress-access-time-update nil
  "Whether to update last-access times on BBDB records.

You should bind this variable, not set it; setting it will break expiry
completely.  It should only be set when changes need to be automatically
made to BBDB records which the user will not care directly about.")

(if (boundp 'bbdb-display-layout-alist)
    (defvar bbdb-expire-autohide-variables
      '((last-access (one-line multi-line pop-up-multi-line))
        (permanent (one-line pop-up-multi-line)))
      "When to hide the special BBDB expiry fields.
A list of mappings from variable names to the layouts to hide them in.

If `omit' is t in a given layout, and the special fields are present
in the order option, no attempt is made to remove them.  It also makes
no attempt to handle omit-specs which are functions (since doing this
in the general case would require solving the halting problem).

This variable only takes effect if set before bbdb-expire is loaded.
(This is a bug.)"))

;;; Code:

;; Keeping things updated.

(defun bbdb-expire-update-last-access-time (record)
  "Update the (non-displayed) `last-access' field in the given RECORD.
(The `last-access' field is used by the `bbdb-expire-record-old-p' predicate.)"
  (let ((bbdb-suppress-changed-records-recording t)
        (bbdb-gag-messages t))
    (or bbdb-expire-suppress-access-time-update
        (bbdb-record-putprop record 'last-access
                             (format-time-string bbdb-time-internal-format
                                                 (current-time))))))

; Call this function whenever a record is noticed.

(add-hook 'bbdb-notice-hook 'bbdb-expire-update-last-access-time)

;; Handling refiling of multiple records containing a `last-access' or
;; `permanent' field; we pick the string which is lexicographically
;; highest. (Here, we see an advantage of the ISO date format;
;; lexicographic ordering is the same as temporal ordering.)

; When we refile them, we want to sort them lexicographically,
; so we need a function that does that.

(defun bbdb-refile-notes-string-most (string1 string2)
  "Returns the string that is not lessp."
  (if (not (string-lessp string1 string2))
      string1
    string2))

; Arrange to use this function when refiling two records that
; contain these fields.

(setq bbdb-refile-notes-generate-alist
      (append '((last-access . bbdb-refile-notes-string-most)
                (permanent . bbdb-refile-notes-string-most))
              bbdb-refile-notes-generate-alist))

;; Apply the requested automatic hides.
;; (In BBDB-2.32, just hide both the variables unconditionally.)

(defun bbdb-expire-auto-edit-bbdb-display-layout-alist ()
  "Automatically edit the `bbdb-display-layout-alist',
applying the transformations in the `bbdb-expire-autohide-variables'.
For BBDB-2.32 and below, merely add the `last-access' and `permanent'
fields to `bbdb-elided-display'."
  (cond
   ((boundp 'bbdb-display-layout-alist)
    (mapc                                 ; Iterate over vars to hide
     #'(lambda (hide-var)
         (mapc                            ; Iterate over layouts to hide them in
          #'(lambda (hide-layout)
              (let ((layout (assq hide-layout bbdb-display-layout-alist)))
                (if layout               ; Existing layout?
                    (let ((omit-option (assq 'omit layout)))
                      (if omit-option    ; Already has an omit clause?
                          (if (and (not (functionp (cdr omit-option)))
                                   (not (memq (car hide-var) omit-option))
                                   (not (eq t (cdr omit-option))))
                              (nconc omit-option (cons (car hide-var) nil)))
                        (nconc layout (cons (list 'omit (car hide-var)) nil))))
                  (setq bbdb-display-layout-alist
                        (nconc bbdb-display-layout-alist
                               (cons (list hide-layout (list 'omit (car hide-var))) nil))))))
          (nth 1 hide-var)))
     bbdb-expire-autohide-variables))
   ;; Primitive, lesser BBDB
   (t
    (setq bbdb-elided-display (nconc bbdb-elided-display '(last-access permanent))))))

;; Initial installation.

(defun bbdb-expire-add-last-access-time ()
  "Add last access times to BBDB records that do not yet have them.
The last access time is set to the current time, because this is safest."
  (interactive)
  ; Do not update last-access timestamps on records that we merely look at.
  (let ((bbdb-suppress-changed-records-recording t)
        (bbdb-gag-messages t)
        (bbdb-expire-suppress-access-time-update t))
    ; For all records, add a `last-access' field set to the current time
    ; if that record does not already have one.
    (mapc #'(lambda (record)
              (or (bbdb-record-getprop record 'last-access)
                  (bbdb-record-putprop record 'last-access
                                       (format-time-string bbdb-time-internal-format
                                                           (current-time)))))
          (bbdb-records))))

(add-hook 'bbdb-expire-load-hook 'bbdb-expire-add-last-access-time)

;; Expiration predicates.

; This is the conventional way of doing expiry.

(defun bbdb-expire-record-old-p (record)
  "Return t if RECORD is older than `bbdb-expire-this-old' days."
  (let ((last-access (bbdb-record-getprop record 'last-access)))
    (and last-access
         (string< last-access
                  (format-time-string bbdb-time-internal-format
                                      (subtract-time (current-time)
                                                     (days-to-time bbdb-expire-this-old)))))))

; Generate all these dull `if-this-field-is-present' predicates with
; a macro.
; (These predicates are used as `bbdb-expire-preservation-functions'.)

(defmacro bbdb-expire-field-foo-p (field)
  "Generate functions that return t if some RECORD contains a given FIELD.
FIELD should be a quoted symbol, just as passed to `bbdb-record-getprop'."
  `(defun ,@(list (intern (concat "bbdb-expire-field-"
                                  (symbol-name (car (cdr field)))
                                  "-p"))) (record)
     ,@(list (concat "Return t if RECORD contains the `"
                     (symbol-name (car (cdr field))) "' field."))
     (bbdb-record-getprop record ,field)))

(bbdb-expire-field-foo-p 'permanent)
(bbdb-expire-field-foo-p 'notes)

;; Expiry.

; FIXME: Is it necessary to use `bbdb-records'? Using it stops this from going
;        on the `bbdb-after-read-db-hook'... :(

(defun bbdb-expire-bbdb ()
  "Expire old records from the BBDB.
Entries are expired if at least one of the `bbdb-expire-expiry-functions'
returns t, and none of the `bbdb-expire-preservation-functions' return t."
  (interactive)
  ; Do not update last-access timestamps on records that we merely look at.
  (let ((bbdb-suppress-changed-records-recording t)
        (bbdb-gag-messages t)
        (bbdb-expire-suppress-access-time-update t))
    ; For all records, if the record should be expired, and fails
    ; all attempts at its preservation, delete it.
    (mapc #'(lambda (record)
              (and (run-hook-with-args-until-success
                    'bbdb-expire-expiry-functions record)
                   (not (run-hook-with-args-until-success
                         'bbdb-expire-preservation-functions record))
                   (bbdb-delete-record-internal record)))
          (bbdb-records))))

(run-hooks 'bbdb-expire-load-hook)

(provide 'bbdb-expire)
